"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = __importDefault(require("axios"));
const bignumber_js_1 = __importDefault(require("bignumber.js"));
const sha3_1 = require("sha3");
const libra_web_account_1 = require("../libra-web-account");
const CursorBuffer_1 = require("../libra-web-core-utils/common/CursorBuffer");
const PathValues_1 = __importDefault(require("../libra-web-core-utils/constants/PathValues"));
const libra_web_net_1 = require("../libra-web-net");
const libra_web_transaction_1 = require("../libra-web-transaction");
var LibraNetwork;
(function (LibraNetwork) {
    LibraNetwork["Testnet"] = "testnet";
    LibraNetwork["Mainnet"] = "mainnet";
    LibraNetwork["Local"] = "local";
})(LibraNetwork = exports.LibraNetwork || (exports.LibraNetwork = {}));
const DefaultFaucetServerHost = 'faucet.testnet.libra.org';
const DefaultProxyServer = process.env.NODE_ENV === 'test'
    ? {
        host: 'localhost',
        network: LibraNetwork.Local,
        port: '8080',
        protocol: 'http',
    }
    : {
        host: 'ac.testnet.trylibra.org',
        network: LibraNetwork.Testnet,
        port: '80',
        protocol: 'http',
    };
class LibraClient {
    constructor(config) {
        this.config = config || DefaultProxyServer;
        const connectionAddress = `${this.config.protocol}://${this.config.host}:${this.config.port}`;
        this.client = new libra_web_net_1.AdmissionControlClient(connectionAddress, null, null);
    }
    /**
     * Fetch the current state of an account.
     *
     *
     * @param {string} address Accounts address
     */
    async getAccountState(address) {
        const result = await this.getAccountStates([address]);
        return result[0];
    }
    /**
     * Fetches the current state of multiple accounts.
     *
     * @param {string[]} addresses Array of users addresses
     */
    async getAccountStates(addresses) {
        for (const address of addresses) {
            if (!libra_web_account_1.AccountAddress.isValidString(address.toHex())) {
                throw new Error(`[${address}] is not a valid address`);
            }
        }
        const request = new libra_web_net_1.UpdateToLatestLedgerRequest();
        addresses.forEach(address => {
            const requestItem = new libra_web_net_1.RequestItem();
            const getAccountStateRequest = new libra_web_net_1.GetAccountStateRequest();
            getAccountStateRequest.setAddress(Uint8Array.from(address.toBytes()));
            requestItem.setGetAccountStateRequest(getAccountStateRequest);
            request.addRequestedItems(requestItem);
        });
        return new Promise((resolve, reject) => {
            this.client.updateToLatestLedger(request, undefined, (error, response) => {
                if (error) {
                    return reject(error);
                }
                resolve(response.getResponseItemsList().map((item, index) => {
                    const stateResponse = item.getGetAccountStateResponse();
                    const stateWithProof = stateResponse.getAccountStateWithProof();
                    if (stateWithProof.hasBlob()) {
                        const stateBlob = stateWithProof.getBlob();
                        const blob = stateBlob.getBlob_asU8();
                        const accountState = this._decodeAccountStateBlob(blob);
                        return accountState;
                    }
                    return libra_web_account_1.AccountState.default(addresses[index]);
                }));
            });
        });
    }
    /**
     * Uses the faucetService on testnet to mint coins to be sent
     * to receiver.
     *
     * Returns the sequence number for the transaction used to mint
     *
     * Note: `numCoins` should be in base unit i.e microlibra (10^6 I believe).
     */
    async mintWithFaucet(receiver, numCoins, waitForConfirmation = true) {
        const serverHost = this.config.faucetServerHost || DefaultFaucetServerHost;
        const coins = new bignumber_js_1.default(numCoins).toString(10);
        const address = receiver.toString();
        const response = await axios_1.default.get(`http://${serverHost}?amount=${coins}&address=${address}`);
        if (response.status !== 200) {
            throw new Error(`Failed to query faucet service. Code: ${response.status}, Err: ${response.data.toString()}`);
        }
        const sequenceNumber = response.data;
        if (waitForConfirmation) {
            await this.waitForConfirmation(libra_web_account_1.AccountAddress.default(), sequenceNumber);
        }
        return sequenceNumber;
    }
    /**
     * Keeps polling the account state of address till sequenceNumber is computed.
     *
     */
    async waitForConfirmation(accountAddress, transactionSequenceNumber) {
        const sequenceNumber = new bignumber_js_1.default(transactionSequenceNumber);
        let maxIterations = 50;
        const poll = (resolve, reject) => {
            setTimeout(() => {
                maxIterations--;
                this.getAccountState(accountAddress)
                    .then(accountState => {
                    if (accountState.sequenceNumber.gte(sequenceNumber)) {
                        return resolve();
                    }
                    if (maxIterations === -1) {
                        reject(new Error(`Confirmation timeout for [${accountAddress.toHex()}]:[${sequenceNumber.toString(10)}]`));
                    }
                    else {
                        poll(resolve, reject);
                    }
                })
                    .catch(reject);
            }, 500);
        };
        return new Promise((resolve, reject) => {
            poll(resolve, reject);
        });
    }
    /**
     * Transfer coins from sender to receipient.
     * numCoins should be in libraCoins based unit.
     *
     * @param sender
     * @param receipientAddress
     * @param numCoins
     */
    async transferCoins(sender, receipientAddress, numCoins) {
        if (typeof receipientAddress === 'string') {
            receipientAddress = libra_web_account_1.AccountAddress.fromHex(receipientAddress);
        }
        const response = await this.execute(libra_web_transaction_1.LibraTransactionFactory.createTransfer(sender.getAddress().toBytes(), Uint8Array.from(receipientAddress.toBytes()), new bignumber_js_1.default(numCoins)), sender);
        return response;
    }
    /**
     * Execute a transaction by sender.
     *
     * @param transaction
     * @param sender
     */
    async execute(transaction, sender) {
        let senderAddress = transaction.sendersAddress;
        if (senderAddress.isDefault()) {
            senderAddress = sender.getAddress();
        }
        let sequenceNumber = transaction.sequenceNumber;
        if (sequenceNumber.isNegative()) {
            const senderAccountState = await this.getAccountState(senderAddress);
            sequenceNumber = senderAccountState.sequenceNumber;
        }
        const rawTransaction = new libra_web_net_1.RawTransaction();
        rawTransaction.setSenderAccount(senderAddress.toBytes());
        rawTransaction.setSequenceNumber(sequenceNumber.toNumber());
        rawTransaction.setProgram(transaction.program);
        rawTransaction.setMaxGasAmount(transaction.gasContraint.maxGasAmount.toNumber());
        rawTransaction.setGasUnitPrice(transaction.gasContraint.gasUnitPrice.toNumber());
        rawTransaction.setExpirationTime(transaction.expirationTime.toNumber());
        const signedTransaction = new libra_web_net_1.SignedTransaction();
        signedTransaction.setSenderPublicKey(sender.keyPair.getPublicKey());
        const sha3 = new sha3_1.SHA3(256);
        // Magic number provided by Libra
        sha3.update('46f174df6ca8de5ad29745f91584bb913e7df8dd162e3e921a5c1d8637c88d16', 'hex');
        sha3.update(new Buffer(rawTransaction.serializeBinary()));
        signedTransaction.setSenderSignature(sender.generateSignature(sha3.digest()));
        signedTransaction.setRawTxnBytes(rawTransaction.serializeBinary());
        const request = new libra_web_net_1.SubmitTransactionRequest();
        request.setSignedTxn(signedTransaction);
        return new Promise((resolve, reject) => {
            this.client.submitTransaction(request, undefined, (error, response) => {
                if (error) {
                    return reject(error);
                }
                resolve(response);
            });
        });
    }
    _decodeAccountStateBlob(blob) {
        const cursor = new CursorBuffer_1.CursorBuffer(blob);
        const blobLen = cursor.read32();
        const state = {};
        for (let i = 0; i < blobLen; i++) {
            const keyLen = cursor.read32();
            const keyBuffer = new Uint8Array(keyLen);
            for (let j = 0; j < keyLen; j++) {
                keyBuffer[j] = cursor.read8();
            }
            const valueLen = cursor.read32();
            const valueBuffer = new Uint8Array(valueLen);
            for (let k = 0; k < valueLen; k++) {
                valueBuffer[k] = cursor.read8();
            }
            state[Buffer.from(keyBuffer).toString('hex')] = valueBuffer;
        }
        return libra_web_account_1.AccountState.from(state[PathValues_1.default.AccountStatePath]);
    }
}
exports.LibraClient = LibraClient;
exports.default = LibraClient;
